
#include "pmacros.h"

#include  <stdio.h>			
#include  <stdlib.h>			
#include  <string.h>			
#include  <pthread.h>			
#include  <sched.h>			

#include  "pool.h"			
#include  "returncode.h"			


#define PTHREAD_POOL_FREE 1
#define PTHREAD_POOL_BUSY 2
#define PTHREAD_POOL_COMPLETED 0


typedef struct pool_thread_argument_t {
	int ego;
	struct pthread_pool_t* a_pool;
} pool_thread_argument_t;



void* pool_thread(void* arg) {
  struct pool_thread_argument_t* a = (pool_thread_argument_t*)arg;
  int status;

  while(1) {

    status=sem_wait(&a->a_pool->waitfor_request[a->ego]);
    if(status!=0)
        {
           returncode("sem_wait pool_thread waitfor",status);
           pthread_exit(NULL);
        }        

    if (a->a_pool->cont==0) 
        pthread_exit(NULL);

    a->a_pool->results[a->ego]=
      a->a_pool->funcs[a->ego](a->a_pool->args[a->ego]);

  status=pthread_mutex_lock(&a->a_pool->sc);
  if(status!=0)
        {
           returncode("pthread_pool_request pthread_mutex_lock ",status);
           return NULL;
        }        

  a->a_pool->state[a->ego]=PTHREAD_POOL_COMPLETED;

  status=pthread_mutex_unlock(&a->a_pool->sc);
  if(status!=0)
        {
           returncode("pthread_pool_request pthread_mutex_unlock ",status);
           return NULL;
        }        

    status=sem_post(&a->a_pool->response_isready[a->ego]);
    if(status!=0)
        {
           returncode("pthread_pool_thread response_ready",status);
           pthread_exit(NULL);
        }        

  }
  return NULL;
}



int pthread_pool_init(struct pthread_pool_t* r, struct pthread_pool_attr_t a) {
  
  int status;
  int i;
  pthread_t tid;
  pthread_attr_t attr;
  struct sched_param param;
  struct pool_thread_argument_t* a_thread_arg;

  r->nb_requests=0;
  r->attr=a;
  r->cont=1;
  for(i=0; i<r->attr.size; i++) 
    {

    r->state[i]=PTHREAD_POOL_FREE;

    status=sem_init(&r->waitfor_request[i],0,0);
    if(status!=0)
        {
           returncode("pthread_pool_init sem_init waitfor_request",status);
           return status;
        }        

    status=sem_init(&r->response_isready[i],0,0);
    if(status!=0)
        {
           returncode("pthread_pool_init sem_init response_isready",status);
           return status;
        }        
    }


  status=pthread_mutex_init(&r->sc, NULL);
  if(status!=0)
        {
           returncode("pthread_pool_request pthread_mutex_lock ",status);
           return status;
        }

  pthread_attr_init(&attr);
  status=pthread_attr_setinheritsched(&attr,PTHREAD_EXPLICIT_SCHED);
  if(status<0)
      {
         returncode("pthread_pool_init pthread_attr_setinheritsched",status);
         return status;
      }        
  status=pthread_attr_setschedpolicy(&attr,SCHED_FIFO);
  if(status<0)
      {
         returncode("pthread_pool_init pthread_attr_setschedpolicy",status);
         return status;
      }        
  param.sched_priority=r->attr.priority;
  status=pthread_attr_setschedparam(&attr,&param);
  if(status<0)
      {
         returncode("pthread_pool_init pthread_attr_setschedparam",status);
         return status;
      }        

  for(i=0; i<r->attr.size; i++) 
    {
    /* allocate thread argument */
    a_thread_arg = (struct pool_thread_argument_t*)malloc(sizeof(struct pool_thread_argument_t));
    a_thread_arg->ego=i;
    a_thread_arg->a_pool=r;

    /* create thread */
    status=pthread_create(&tid,&attr,pool_thread,a_thread_arg);
    if(status<0)
        {
           returncode("pthread_pool_start pthread_create",status);
           return status;
        }        
    }

    return 0;
}



int pthread_pool_request(struct pthread_pool_t* r, request_func_t func, void* arg, pthread_pool_request_t* id) {

  int status, i;

  status=pthread_mutex_lock(&r->sc);
  if(status!=0)
        {
           returncode("pthread_pool_request pthread_mutex_lock ",status);
           return status;
        }        


  if (r->nb_requests==r->attr.size)
    {
       status=pthread_mutex_unlock(&r->sc);
       if(status!=0)
           {
           returncode("pthread_pool_request pthread_mutex_lock ",status);
           return status;
           }        
       errno=EBUSY;
       return -1;
    }

  for(i=0; i<r->attr.size; i++) 
    {
    if(r->state[i]==PTHREAD_POOL_FREE)
       {
       *id=i;
       r->nb_requests++;
       r->state[i]=PTHREAD_POOL_BUSY;
       r->funcs[i]=func;
       r->args[i]=arg;
       status=sem_post(&r->waitfor_request[i]);
       if(status!=0)
           {
           returncode("pthread_pool_request sem_post waitfor",status);
           return status;
           }        
       status=pthread_mutex_unlock(&r->sc);
       if(status!=0)
           {
           returncode("pthread_pool_request pthread_mutex_lock ",status);
           return status;
           }        
       return 0;
       }
    }

  status=pthread_mutex_unlock(&r->sc);
  if(status!=0)
        {
           returncode("pthread_pool_request pthread_mutex_lock ",status);
           return status;
        }        


  return 0;
}


int pthread_request_status(struct pthread_pool_t* r, pthread_pool_request_t id) {

  int status;
  int result;

  status=pthread_mutex_lock(&r->sc);
  if(status!=0)
        {
           returncode("pthread_pool_thread pthread_mutex_lock ",status);
           return 0;
        }        

  result=r->state[id];

  status=pthread_mutex_unlock(&r->sc);
  if(status!=0)
        {
           returncode("pthread_pool_thread pthread_mutex_unlock ",status);
           return 0;
        }        

  if (result!=PTHREAD_POOL_COMPLETED)
	  return -1;
  else  return 0;
}


void* pthread_pool_response(struct pthread_pool_t* r, pthread_pool_request_t id) {

  int status;
  void* result;

  status=sem_wait(&r->response_isready[id]);
  if(status!=0)
      {
      returncode("pthread_pool_response sem_wait response_ready",status);
      return NULL;
      }        

  status=pthread_mutex_lock(&r->sc);
  if(status!=0)
        {
           returncode("pthread_pool_thread pthread_mutex_lock ",status);
           return NULL;
        }        

  r->nb_requests--;
  result=r->results[id];
  r->state[id]=PTHREAD_POOL_FREE;

  status=pthread_mutex_unlock(&r->sc);
  if(status!=0)
        {
           returncode("pthread_pool_thread pthread_mutex_unlock ",status);
           return NULL;
        }        

  return result;
}


int pthread_pool_destroy(struct pthread_pool_t* r) {

  int i;
  int status;

  r->cont=0;
  for(i=0; i<r->attr.size; i++) 
    {
    status=sem_post(&r->waitfor_request[i]);
    if(status!=0)
        {
           returncode("pthread_pool_destroy sem_post",status);
           return status;
        }        
    }

  return 0;
}




int pthread_pool_attr_destroy(struct pthread_pool_attr_t *a) {
   return 0;
}


int pthread_pool_attr_init(struct pthread_pool_attr_t *a) {
   a->size=3;
   a->priority=100;
   return 0;
}


int pthread_pool_attr_setpriority(struct pthread_pool_attr_t *a, int prio) {
   a->priority=prio;
   return 0;
}


int pthread_pool_attr_getpriority(struct pthread_pool_attr_t a) {
  return a.priority;
}

int pthread_pool_attr_setsize(struct pthread_pool_attr_t *a, int s) {
   a->size=s;
   return 0;
}

int pthread_pool_attr_getsize(struct pthread_pool_attr_t a) {
  return a.size;
}





